/*
 * uid.h:
 * User and group handling.
 *
 * Copyright (c) 2015 David Suárez.
 * Email: david.sephirot@gmail.com
 *
 */

#ifndef __UID_H__
#define __UID_H__

void drop_root(const char *username);

#endif /* __UID_H__ */
