/*
 * driftnet.h:
 * Pick out images from passing network traffic.
 *
 * Copyright (c) 2001 Chris Lightfoot.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 * $Id: driftnet.h,v 1.13 2004/04/08 23:06:29 chris Exp $
 *
 */

#ifndef __DRIFTNET_H_ /* include guard */
#define __DRIFTNET_H_

#ifdef HAVE_CONFIG_H
    #include <config.h>
#endif

#define PROGNAME DRIFTNET_PROGNAME

#endif /* __DRIFTNET_H_ */
