/*
 * web_data.h:
 * static content files list.
 *
 * Copyright (c) 2012-2019 David Suárez.
 * Email: david.sephirot@gmail.com
 *
 */

#ifndef __WEB_DATA_H_
#define __WEB_DATA_H_

#define STATIC_WEB_DIRECTORY STATIC_WEB_DIR

extern char* static_files[];

#endif /* __WEB_DATA_H_ */
